"""
Date: 9/6/2019 4:45 PM
Author: Achini
"""

import gensim
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import pandas as pd
from sklearn.manifold import TSNE
from sklearn.metrics.pairwise import cosine_similarity
from scipy import spatial


def similarity_between_emotions(model):
    """
    Function to calculate the similarity metrics between emotions and create a heatmap
    :param model: trained model
    :return: None
    """
    csv = []
    headings = ['emotion', 'target', 'value']
    csv.append(headings)
    for e1 in EMOTIONS:
        for e2 in EMOTIONS:
            s = model.wv.similarity(e1, e2)
            csv.append([e1, e2, s])

    pd.DataFrame(csv).to_csv(RESULT_ROOT + 'heatmap.csv')


def tsne_plot_similar_words(title, labels, embedding_clusters, word_clusters, a, filename):
    """
    Function to visualize the related terms in 2d space using tsne, and to write the related terms to a csv
    :param title: title of the image
    :param labels: keys (emotions)
    :param embedding_clusters: embedding_clusters from the we model
    :param word_clusters: word_clusters from the we model
    :param a: alpha for the plot.annotate function
    :param filename: filename for the figure
    :return: None
    """
    plt.figure(figsize=(16, 9))
    colors = cm.rainbow(np.linspace(0, 1, len(labels)))
    for label, embeddings, words, color in zip(labels, embedding_clusters, word_clusters, colors):
        csv = []
        x = embeddings[:, 0]
        y = embeddings[:, 1]
        plt.scatter(x, y, c=color, alpha=a, label=label)
        for i, word in enumerate(words):
            plt.annotate(word, alpha=0.9, xy=(x[i], y[i]), xytext=(5, 2),
                         textcoords='offset points', ha='right', va='bottom', size=8)
            csv.append([label, word])
        """ write output to csv"""
        pd.DataFrame(csv).to_csv(RESULT_ROOT + label + '.csv')
    plt.legend(loc=4)
    plt.title(title)
    plt.grid(False)

    plt.savefig(filename, format='png', dpi=1200, bbox_inches='tight')
    plt.show()
    pd.DataFrame(csv).to_csv(RESULT_ROOT + 'model_outcomes.csv')


def query_similar_terms(model, keys, figure_name):
    """
    Function to query similar terms for a set of emotions
    :param model: trained word2vec model
    :param keys: words to be queried
    :return: None
    """

    embedding_clusters = []
    word_clusters = []
    for word in keys:
        embeddings = []
        words = []
        for similar_word, _ in model.most_similar(word, topn=30):
            words.append(similar_word)
            embeddings.append(model[similar_word])
        embedding_clusters.append(embeddings)
        word_clusters.append(words)

    embedding_clusters = np.array(embedding_clusters)
    n, m, k = embedding_clusters.shape
    tsne_model_en_2d = TSNE(perplexity=15, n_components=2, init='pca', n_iter=3500, random_state=32)
    embeddings_en_2d = np.array(tsne_model_en_2d.fit_transform(embedding_clusters.reshape(n * m, k))).reshape(n, m, 2)

    tsne_plot_similar_words('Similar words', keys, embeddings_en_2d, word_clusters, 0.7,
                             figure_name)


if __name__ == '__main__':
    model = gensim.models.Word2Vec.load('model_name')
    RESULT_ROOT = 'path to save output'
    figure_name = RESULT_ROOT + 'sample.png'

    EMOTIONS = ['anger', 'anticipation', 'disgust', 'fear', 'joy', 'sad', 'surprise', 'trust']

    """ tsne """
    query_similar_terms(model, EMOTIONS, figure_name)

    """ similarity """
    compute_similarity_two_vectors(MODEL)


# imports needed and set up logging
import gensim
import logging
import pandas as pd
from nltk.tokenize import word_tokenize
import string
from ..utils import text_processor as text_utils


def create_wordembedding_model(posts, model_name, test_word):
    """
    Function to create a word2vec embedding model
    :param posts: training data
    :param model_name: name to save the word embedding model
    :param test_word: test word to query the trained model
    :return: trained word embedding model
    """
    documents = []
    for post in posts:
        documents.append(word_tokenize(post))
    print('data processing completed')

    """ create word embedding model """
    model = gensim.models.Word2Vec(documents, size=100, window=10, min_count=2, workers=10)
    model.train(documents, total_examples=len(documents), epochs=10)

    print('-----training complete--------')

    model.save(model_name)
    print('-----model saved-------------')

    w1 = test_word
    print(model.wv.most_similar(positive=w1))

    return model


if __name__ == '__main__':

    data_path = 'enter data path here'
    model_name = 'we_emotions.model'
    test_word = 'sad'

    df = pd.read_csv(data_path)
    posts = df['post'].apply(text_utils.clean_text)
    word2vec_model = create_wordembedding_model(posts, model_name, test_word)



"""
Date: 2/28/2020 6:47 PM
Author: Achini
"""
from ...core.emotions import emotion_extractor as emotions
from statistics import mean


def get_emotion_engagement_score(posts_list, EMO_RESOURCES):
    """
    Function to calculate emotion enagement
    :param posts_list: post list of a user
    :return: emotion_engagement_score
    """
    """ get emotion-less posts """
    emotion_values = []
    all_emotion_profiles = []
    emotion_less_posts = 0
    highly_emotional_posts = 0

    for post in posts_list:
        emotion_profile = emotions.get_emotion_profile_per_post(post, EMO_RESOURCES)
        if sum(emotion_profile.values()) == 0:
            emotion_less_posts += 1

        for val in emotion_profile.values():
            emotion_values.append(val)

        all_emotion_profiles.append(emotion_profile.values())

    emotion_mean = mean(emotion_values)

    """ get highly emotional posts """
    for emo_value_profile in all_emotion_profiles:
        if sum(emo_value_profile) > emotion_mean:
            highly_emotional_posts += 1

    if emotion_less_posts > 0:
        emotion_engagement_score = highly_emotional_posts / emotion_less_posts

    else:
        emotion_engagement_score = highly_emotional_posts

    return emotion_engagement_score


def get_participation_score(user_posts, all_posts, num_all_users):
    """
    Function to calculate the relative participation score
    :param user_posts: (list) posts of a user
    :param all_posts: (list) all posts
    :param num_all_users: (int) number of users
    :return: (number) participation score
    """
    user_post_len = 0
    all_post_len = 0
    num_user_posts = len(user_posts)
    num_all_posts = len(all_posts)

    """ calculate average post length of a user """
    for u_post in user_posts:
        user_post_len += len(u_post.split())
    if num_user_posts > 0:
        avg_post_len_USER = user_post_len / num_user_posts
    else:
        avg_post_len_USER = 0

    """ calculate average post length of the group """
    for post in all_posts:
        all_post_len += len(post.split())
    if num_all_posts > 0:
        avg_post_len_ALL = all_post_len / num_all_posts
    else:
        avg_post_len_ALL = 0

    """ average length of post - relative measure """
    if avg_post_len_ALL > 0:
        avg_post_len = avg_post_len_USER / avg_post_len_ALL
    else:
        avg_post_len = 0

    """ average number of posts - relative measure """
    if (num_all_posts / num_all_users) > 0:
        avg_posts_by_user = num_user_posts / (num_all_posts / num_all_users)
    else:
        avg_posts_by_user = 0

    """ overall participation score """
    participation_score = avg_post_len * avg_posts_by_user

    return participation_score



"""
Date: 2/28/2020 10:51 AM
Author: Achini
"""

import re


def get_physical_sym_profile(posts, EMO_RESOURCES):
    """
    Function to extract mentions of physical symptoms
    :param post: clean post
    :return: dictionary containing physical symptom score (number) and mentions (list)
    """
    PHYSICAL_MAP = EMO_RESOURCES['PHYSICAL']
    num_posts = len(posts)
    physical_terms = {}
    if type(posts) is str:
        posts = [posts]
    for post in posts:
        try:
            sent_words = str(post).lower().split()
            for term in PHYSICAL_MAP:
                term = term.strip().lower()

                if len(term.split()) > 1:
                    pattern = re.compile(r'\b%s\b' % term, re.I)
                    match_found = pattern.search(post)
                else:
                    match_found = term in sent_words
                if match_found:
                    if term in physical_terms.keys():
                        physical_terms[term] = physical_terms[term] + 1
                    else:
                        physical_terms[term] = 1
        except:
            match_found = False

    if num_posts > 0:
        for key in physical_terms.keys():
            physical_terms[key] = physical_terms[key] / num_posts

    return physical_terms



"""
Date: 2/27/2020 5:59 PM
Author: Achini
"""


def check_negation(text, NEGATION_MAP):
    """
    Utility function to check negation of an emotion
    :param text: text chunk with the emotion term
    :return: boolean value for negation
    """
    neg_word_list = NEGATION_MAP
    neg_match = False
    for neg_word in neg_word_list:
        if neg_word.strip() in text:
            neg_match = True

    return neg_match



def check_intensifiers(text, INTENSIFIER_MAP):
    """
    Utility function to check intensifiers of an emotion
    :param text: text chunk with the emotion term
    :return: boolean value and booster value for intensifiers
    """
    # BOOSTER_MAP = {"B_INCR": 2,
    #                "B_DECR": 0.5}
    intensity_word_list = INTENSIFIER_MAP
    has_intensity = False
    booster = 'NULL'
    for int_term in intensity_word_list:
        intensifier = int_term.split(':')[0].strip()

        if intensifier in text:
            has_intensity = True
            booster = float(int_term.split(':')[2].strip())

    return has_intensity, booster


def get_opposite_emotion(key):
    """
    Utility function to get the opposite emotion of a given emotion
    :param key: emotion to be processed
    :return: opposite emotion, None if no opposite emotion is found
    """
    opposite_emotions = {"joy": "anger",
                         "sad": "joy",
                         "anticipation": "anger",
                         "trust": "fear",
                         'fear': 'trust',
                         'anger' : 'joy',
                         'afraid': 'trust'
                         }

    if opposite_emotions.keys().__contains__(key):
        return opposite_emotions[key]
    else:
        return None


def get_sentiment_of_emotions(emotion):
    """
    Utility function to get the POS/NEG categorization of an emotion
    :param emotion: emotion to be processed
    :return: POS, NEG category
    """
    POS = ['joy', 'trust', 'anticipation', 'surprise']
    NEG = ['sad', 'fear', 'disgust', 'anger', 'hopelessness', 'loneliness', 'distress']

    if emotion in POS:
        return 'POS'
    elif emotion in NEG:
        return 'NEG'
    else:
        return None


def get_empathetic_templates():
    templates = ['sorry to', 'sad to', 'sorry for', 'must be hard', 'sorry that', 'for you', 'QUE']
    return templates


def is_firstperson_post(sent_words):
    """
    Resolve pro noun use
    :param sent_words: sentence
    :return: first person post True/False
    """
    first_person_list = ['i', 'myself', 'me']
    second_person_list = ['you', 'yourself', 'your']
    i_count = 0
    you_count = 0
    match = False
    for term1 in sent_words:
        for i in first_person_list:
            if term1 == i:
                i_count += 1

    for term2 in sent_words:
        for j in second_person_list:
            if term2 == j:
                you_count += 1

    if i_count >= you_count:
        match = True
    return match


"""
Date: 2/27/2020 5:58 PM
Author: Achini
"""

import re
from . import emotion_utilities as emotion_utils
from nltk.tokenize import sent_tokenize
from ...utils import text_processor as text_utils
import operator


def get_emotion_profile_per_post(clean_post, EMO_RESOURCES):
    """
    Function to process each post to extract emotion mentions
    :param clean_post: post to be processed
    :return emotion_profile (dictionary): dictionary containing the intensity for each emotion
    """
    EMOTION_MAP = EMO_RESOURCES['EMOTIONS']
    NEGATION_MAP = EMO_RESOURCES['NEGATION']
    INTENSIFIER_MAP = EMO_RESOURCES['INTENSIFIERS']

    emotion_profile = {}
    emo_counts = {}
    all_emo_words = []
    emo_seq = []

    sent_words = str(clean_post).lower().split()
    post_len = len(sent_words)
    """ emotion extraction """
    # try:
    for key in EMOTION_MAP.keys():
        emotion_val_list = []
        for emoWord in EMOTION_MAP[key]:
            try:
                emoWord = emoWord.strip().lower()
                matched_emotion = emoWord.strip().lower()

                if len(emoWord.split()) > 1:
                    pattern = re.compile(r'\b%s\b' % emoWord, re.I)
                    match_found = pattern.search(clean_post)
                else:
                    match_found = emoWord in sent_words

                if match_found:
                    selected_emo = key
                    all_emo_words.append(emoWord)

                    """ handle intensifiers"""
                    if len(emoWord.split()) > 1:
                        emoWord = emoWord.split()[0]
                    end_ind_int = sent_words.index(emoWord)
                    start_ind_int = end_ind_int - 3
                    if start_ind_int < 0:
                        start_ind_int = 0
                    text_chunk_int = sent_words[start_ind_int:end_ind_int]
                    has_intensity, booster = emotion_utils.check_intensifiers(text_chunk_int, INTENSIFIER_MAP)

                    """ handle negation """
                    end_ind = sent_words.index(emoWord)
                    if end_ind < 3:
                        start_ind = 0
                    else:
                        start_ind = end_ind - 3
                    text_chunk = sent_words[start_ind:end_ind]
                    negation = emotion_utils.check_negation(text_chunk, NEGATION_MAP)
                    if negation:
                        # get opposite emotion
                        selected_emo = emotion_utils.get_opposite_emotion(key)

                    """ handle empathy """
                    if key == 'sad':
                        templates = emotion_utils.get_empathetic_templates()
                        for temp in templates:
                            pattern = re.compile(r'\b%s\b' % temp, re.I)
                            empath_match = pattern.search(clean_post)
                            if empath_match is not None:
                                selected_emo = None
                            else:
                                firstperson = emotion_utils.is_firstperson_post(sent_words)
                                if not firstperson:
                                    selected_emo = None

                    """ assign value """
                    if selected_emo is not None:
                        emo_value = 1
                        emo_seq.append(selected_emo)
                        if has_intensity:
                            emo_value = 1 * booster
                        else:
                            emo_value = 1 * 0.5
                    else:
                        emo_value = 0

                    emotion_val_list.append(emo_value)
            except Exception as e:
                print(e)

        emotion_profile[key] = sum(emotion_val_list)

    return emotion_profile, emo_seq


def get_emotion_sequence(posts, EMO_RESOURCES):
    emo_seq = []
    agg_profile = {}

    for post in posts:
        clean_post = text_utils.clean_text(post)
        sentences = sent_tokenize(clean_post)

        for sent in sentences:
            profile, seq = get_emotion_profile_per_post(sent, EMO_RESOURCES)
            if profile != {}:
                max_emotion = max(profile.items(), key=operator.itemgetter(1))[0]
                max_emo_val = max(profile.items(), key=operator.itemgetter(1))[1]

            if max_emo_val > 0:
                for key in profile:
                    if profile[key] == max_emo_val:
                        emo_seq.append(max_emotion)
                    if key in agg_profile.keys():
                        agg_profile[key] = agg_profile[key] + profile[key]
                    else:
                        agg_profile[key] = profile[key]

    return emo_seq, agg_profile


def get_aggregated_emotions(all_posts, EMO_RESOURCES):
    """
    Function to get the aggregated emotions
    :param all_posts: all posts in a session/user
    :return: (dictionary) agg_profile with intensities for granular emotions,
             (dictionary) agg_sum with aggregated POS/NEG intensities
    """
    agg_profile = {}
    agg_POS = 0.0
    agg_NEG = 0.0
    for KEY in EMO_RESOURCES['EMOTIONS'].keys():
        agg_profile[KEY] = 0.0

    for post in all_posts:
        try:
            post_profile = get_emotion_profile_per_post(post, EMO_RESOURCES)
            for e in post_profile:
                agg_profile[e] = agg_profile[e] + post_profile[e]

                if emotion_utils.get_sentiment_of_emotions(e) == 'POS':
                    agg_POS += post_profile[e]
                elif emotion_utils.get_sentiment_of_emotions(e) == 'NEG':
                    agg_NEG += post_profile[e]
        except:
            post_profile = {}

    agg_sum = {'aggPOS': agg_POS, 'aggNEG': agg_NEG}

    return agg_profile, agg_sum



"""
Date: 3/4/2020 11:55 AM
Author: Achini
"""

import pickle, os
from .core.emotions import emotion_extractor as emotion_extractor
from .utils import text_processor as text_utils
from .core.summary import keyphrase_extractor as keyphrase_extractor
from .core.clinical_info import clinical_info_extractor as clinical_info_extractor


def load_emotion_dictionaries():
    with open(os.path.dirname(__file__)+'/models/emotions/emotions_plutchik.pkl', 'rb') as f:
        EMOTION_MAP = pickle.load(f)
    with open(os.path.dirname(__file__)+'/models/emotions/intensifier_vocab_v2.pkl', 'rb') as f:
        INTENSIFIER_MAP = pickle.load(f)
    with open(os.path.dirname(__file__)+'/models/emotions/negation_vocab_v2.pkl', 'rb') as f:
        NEGATION_MAP = pickle.load(f)
    with open(os.path.dirname(__file__)+'/models/clinical_info/physical.pkl', 'rb') as f:
        PHYSICAL = pickle.load(f)

    EMO_RESOURCES = {'EMOTIONS': EMOTION_MAP,
                     'NEGATION': NEGATION_MAP,
                     'INTENSIFIERS': INTENSIFIER_MAP,
                     'PHYSICAL': PHYSICAL}

    return EMO_RESOURCES


EMO_RESOURCES = load_emotion_dictionaries()
if __name__ == '__main__':
    text_1 = 'still i feel very sad about the unexpected incident. hopefully the pain will be less, and i am grateful for..'
    clean_text_1 = text_utils.clean_text(text_1)
    emotion_profile, emo_seq = emotion_extractor.get_emotion_profile_per_post(clean_text_1, EMO_RESOURCES)
    clinical_info = clinical_info_extractor.get_physical_sym_profile(clean_text_1, EMO_RESOURCES)
    keyphrases = keyphrase_extractor.analyze_keyphrases(clean_text_1)

    print(keyphrases)
    print(clinical_info)
    print(emotion_profile)

def get_emotion_profile_per_post(post):
    return emotion_extractor.get_emotion_profile_per_post(text_utils.clean_text(post), EMO_RESOURCES)
def get_physical_sym_profile(post):
    return clinical_info_extractor.get_physical_sym_profile(text_utils.clean_text(post), EMO_RESOURCES)
def analyze_keyphrases(post):
    return keyphrase_extractor.analyze_keyphrases(text_utils.clean_text(post))



from .emotion import get_emotion_profile_per_post as get_emotion_profile_per_post
from .emotion import get_physical_sym_profile as get_physical_sym_profile
from .emotion import analyze_keyphrases as analyze_keyphrases


"""
Date: 2/27/2020 6:01 PM
Author: Achini
"""
import re
import string


""" define custom templates to ignore in posts"""
TEMPLATES = ['good morning', 'good afternoon', 'good night', 'good evening']


def clean_text(post):
    """
    Function to filter basic greetings and clean the input text.
    :param post: raw post
    :return: clean_post or None if the string is empty after cleaning
    """
    post = str(post)

    """ filtering basic greetings """
    for template in TEMPLATES:
        if template in str(post).lower():
            post = post.replace(template, '')

    """ clean text """
    raw_text = str(post).replace('\'', ' ')
    translator = re.compile('[%s]' % re.escape(string.punctuation))
    clean_text_sub = translator.sub(' ', raw_text)
    clean_text = re.sub(' +', ' ', clean_text_sub).strip()

    if clean_text == 'nan' or clean_text is None:
        return ''
    else:
        return clean_text



"""
Date: 2/29/2020 8:51 PM
Author: Achini
"""
from ...core.summary import rake as RAKE


def combine_posts(posts):
    """
    Function to combine posts to single text
    :param posts: all posts by a user
    :return: (String) combined text
    """
    combined_text = '.'.join(posts)
    return combined_text


def analyze_keyphrases(posts):
    """
    Function to extract key phrases in a set of posts, done using RAKE (Rose, S., D. Engel, N. Cramer, and W. Cowley (2010).)
    :param posts: all posts of a user
    :return: (dictionary) high frequent key phrases
    """
    if type(posts) is not str:
        text = combine_posts(posts)
    else:
        text = posts
    keywords = RAKE.rake.run(text)
    high_freq_keywords = []
    for pair in keywords:
        """ set threshold to select keyphrases """
        try:
            if pair[1] > 3.5:
                high_freq_keywords.append([pair[0], pair[1]])
        except IndexError:
            continue
    return high_freq_keywords
